/* global browser, Port, cfg:true, gen_nodes, media_type */
"use strict";

var init = function(win, doc, head) {
	// cfg is not available in normal web-pages, but in extension pages only
	if (typeof Port === "undefined" || typeof cfg === "object" || !doc || !doc.body) {
		return;
	}

	var pdsp = function(e, d, p) {
		if (!e || !e.preventDefault || !e.stopPropagation) {
			return;
		}

		if (d === void 0 || d === true) {
			e.preventDefault();
		}

		if (p === false) {
			return;
		}

		if (e.stopImmediatePropagation) {
			e.stopImmediatePropagation();
		}
		else {
			e.stopPropagation();
		}
	},
	shortcut = {
		specKeys: {
			8: "Backspace", 9: "Tab", 13: "Enter",
			16: "shift", 17: "ctrl", 18: "alt", // control keys are not capital
			27: "Esc", 32: "Space",
			33: "PgUp", 34: "PgDn", 35: "End", 36: "Home",
			37: "Left", 38: "Up", 39: "Right", 40: "Down",
			45: "Ins", 46: "Del",
			96: "0", 97: "1", 98: "2", 99: "3", 100: "4",
			101: "5", 102: "6", 103: "7", 104: "8", 105: "9",
			106: "*", 107: "+", 109: "-", 110: ".", 111: "/",
			186: ";", 187: "=", 188: ",", 189: "-", 190: ".",
			191: "/", 192: "`", 219: "[", 220: "\\", 221: "]", 222: "'",
			112: "F1", 113: "F2", 114: "F3", 115: "F4", 116: "F5", 117: "F6",
			118: "F7", 119: "F8", 120: "F9", 121: "F10", 122: "F11", 123: "F12"
		},
		isModifier: function(e) {
			return e.which > 15 && e.which < 19;
		},
		key: function(e) {
			return this.specKeys[e.which] || String.fromCharCode(e.which).toUpperCase();
		}
	};

	cfg = JSON.parse((head && (head.data || head.message || head)).prefs);

	var media = doc.body.querySelector("img" + (cfg.video ? ", video" : ""));

	if (!media) {
		init = null;
		return;
	}

	media.style.display = "none";

	if (browser.opera) {
		if (media_type() === "img") {
			win.donotrun = media.error = true;
			media.naturalWidth = 0;

			media.onclick =
				win.ondragstart =
				win.onkeypress =
				win.onmousedown =
				win.onmouseup =
				win.onmousemove = null;
		}
		else {
			doc.body.replaceChild(media, media.parentNode);
		}
	}

	["class", "style"].forEach(function(attr) {
		doc.documentElement.removeAttribute(attr);
		doc.body.removeAttribute(attr);
	});

	var root, init_params,
		after_calc,
		moving,
		winW, winH,
		sX, sY,
		noFit = {cur: false, real: false},
		lastEvent = {},
		progress = null,
		cancelAction = false,
		fZoom = null,
		dragSlide = [],
		borderSize = 0,
		MAXSIZE = 0x7fff;

	if (head = doc.querySelector("head")) {
		doc.documentElement.removeChild(head);
	}

	head = doc.createElement("head");

	if (cfg.favicon) {
		root = doc.createElement("link");
		root.rel = "shortcut icon";
		root.href = cfg.favicon === "%url" && media_type() === "img"
			? win.location.href
			: cfg.favicon.replace("%url", encodeURIComponent(win.location.href));
	}

	if (root) {
		head.appendChild(root);
	}

	root = doc.createElement("style");
	root.appendChild(doc.createTextNode([
		"html, body{width: 100%; height: 100%; -webkit-user-select: none; -moz-user-select: none; -ms-user-select: none}",
		"body {cursor: default; margin:0; padding: 0; font: 12px 'Trebuchet MS', sans-serif}",
		"body > img, body > video {margin:0 !important; background-clip: padding-box;" + browser["box-sizing_css"] + ": border-box; position: absolute}",
		"h2 {margin: 0; padding-top: 15%; text-align: center; text-shadow: 1px 1px 15px #000; color: #fff; font: 700 30px 'Trebuchet MS', sans-serif}",
		".m_1 {max-width: 100%; max-height: 100%; width: auto; height: auto}",
		".m_2 {width: 100%; height:auto} .m_3 {height: 100%; width: auto}",
		"a {text-decoration: none}",
		":root.error img, :root.error video {box-shadow: 0 0 10px red}",
		":root.error>body>h2 {text-shadow: 1px 1px 15px firebrick}",
		":focus {outline: 0}",
		"#menu {position: fixed; " + browser["transition_css"] + ": opacity .25s, left .2s; width: 50px; height: 33.33%; top: 0; opacity: 0}",
		"ul {display: inline-block; text-align: center; margin: 0; padding: 5px 15px; list-style: none; font-size: 25px; font-weight: 700; background: rgba(0, 0, 0, .6); color: #fff}",
		"li {display: block; position: relative}",
		"li[data-cmd] {cursor: pointer}",
		"li[data-cmd=cycle], li[data-cmd=rotate] {position: relative}",
		"li[data-cmd=cycle]::after {content: '\u2195'; position: absolute; left: 25%}",
		"li[data-cmd=rotate]::after {content: '\u21BB'; position: absolute; left: 2px}",
		"li ul {position: absolute; left: 100%; top: 0; visibility: hidden; opacity: 0; " + browser["transition_css"] + ": visibility .4s, opacity .2s .3s; text-align: left}",
		"li:hover ul {display: block; opacity: 1; visibility: visible}",
		"li ul>li {display: block !important; font-size: 15px; padding: 3px 0}",
		".send_hosts li > a {color: #fff; background-size: 16px 16px; background-repeat: no-repeat; background-position: 0 3px; padding-left: 25px;}",
		"#menu>ul>li:hover, .send_hosts li > a:hover {color: silver}",
		"li ul>li {white-space: nowrap}",
		".filters>ul {font-size: 15px} .filters>form {margin: 0; padding: 0}",
		"body.frames, body.frames {text-align: center; overflow-y: scroll !important} body.frames * {margin: 2px}",
		"#frames > canvas, #frames > img {display: none}",
		"#current_frame+output {display: inline-block; font-family: Conoslas, monospace; width: 65px; text-align: right}",
		".back {font-size: 150%; color: black; text-decoration: none}",
		"#frames.showall > img, #frames.showall > canvas {display: inline-block !important}",
		(cfg.customCSS || "body {background: silver; overflow: hidden} img,video {background: white; box-shadow: 0 0 5px grey}")
	].join("")));

	head.appendChild(root);
	doc.documentElement.insertBefore(head, doc.body);

	init = function() {
		if (!(media.naturalWidth || media.width || media.videoWidth || init_params === null)) {
			if (init_params && ++init_params.loop >= init_params.maxLoop && progress) {
				clearInterval(progress);
			}

			return;
		}

		if (progress) {
			clearInterval(progress);
		}

		root.classList.add(media_type());

		if (media_type() === "video") {
			media.naturalWidth = media.videoWidth;
			media.naturalHeight = media.videoHeight;
		}
		else {
			doc.body.replaceChild(media, doc.images[0]);
		}

		[].forEach.call(doc.body.childNodes, function(node) {
			if (media !== node) {
				node.parentNode.removeChild(node);
			}
		});

		media.style.display = "block";

		if (!cfg.imgInfo) {
			cfg.imgInfo = false;
		}

		head.appendChild(doc.createElement("style")).textContent =
			"body > img, body > video {max-width:" +
				Math.min(
						MAXSIZE,
						Math.floor(media.naturalWidth * MAXSIZE / media.naturalHeight)
					) +
				"px;max-height:" +
				Math.min(
						MAXSIZE,
						Math.floor(media.naturalHeight * MAXSIZE / media.naturalWidth)
					) + "px}";

		cfg.hiddenScrollbars = win.getComputedStyle(root).overflow === "hidden"
										|| win.getComputedStyle(doc.body).overflow === "hidden";

		var menu = doc.body.appendChild(doc.createElement("div"));
		menu.id = "menu";

		if (win.getComputedStyle(menu).display === "none") {
			doc.body.removeChild(menu);
			menu = null;
		}
		else {
			menu.style.cssText = "-webkit-filter: blur(1px); filter: blur(1px)";

			if (menu.style.webkitFilter || menu.style.filter) {
				media.filters = {};
				menu.onchange = function(e) {
					var t = e.target, f = "";

					if (t.value) {
						if (t.value === t.defaultValue) {
							delete media.filters[e.target.parentNode.textContent.trim()];
						}
						else {
							media.filters[t.parentNode.textContent.trim()] = t.value + t.getAttribute("unit");
						}

						for (var fname in media.filters) {
							f += fname + "(" + media.filters[fname] + ") ";
						}
					}

					media.style.webkitFilter = f;
					media.style.filter = f;
				};
			}

			gen_nodes(menu.appendChild(doc.createElement("ul")), [
				{tag: "li", attrs: {"data-cmd": "cycle"}, text: "\u2194"},
				{tag: "li", attrs: {"data-cmd": "zoom"}, nodes: [
					{tag: "span", attrs: {style: "visibility: hidden"}, text: "."},
					{tag: "span", attrs: {style: "transform: rotate(-45deg); position: absolute; left: 20%; font-size: 125%"}, text: "\u26B2"}
				]},
				{tag: "li", attrs: {"data-cmd": "flip"}, text: "\u21CB"},
				{tag: "li", attrs: {"data-cmd": "rotate"}, text: "\u21BA"},
				media.filters ? {tag: "li", attrs: {"class": "filters"}, nodes: [
					"\u2261",
					{tag: "form", nodes: [{tag: "ul", nodes: [
						{tag: "li", nodes: [{tag: "input", attrs: {type: "range", min: 0, max: 250, step: 10, value: 100, unit:"%"}}, " brightness"]},
						{tag: "li", nodes: [{tag: "input", attrs: {type: "range", min: 0, max: 300, step: 25, value: 100, unit:"%"}}, " contrast"]},
						{tag: "li", nodes: [{tag: "input", attrs: {type: "range", min: 0, max: 1000, step: 50, value: 100, unit:"%"}}, " saturate"]},
						{tag: "li", nodes: [{tag: "input", attrs: {type: "range", min: 0, max: 100, step: 25, value: 0, unit:"%"}}, " grayscale"]},
						{tag: "li", nodes: [{tag: "input", attrs: {type: "range", min: 0, max: 100, step: 100, value: 0, unit:"%"}}, " invert"]},
						{tag: "li", nodes: [{tag: "input", attrs: {type: "range", min: 0, max: 100, step: 20, value: 0, unit:"%"}}, " sepia"]},
						{tag: "li", nodes: [{tag: "input", attrs: {type: "range", min: 0, max: 360, step: 36, value: 0, unit:"deg"}}, " hue-rotate"]},
						{tag: "li", nodes: [{tag: "input", attrs: {type: "range", min: 0, max: 20, step: 1, value: 0, unit:"px"}}, " blur"]}
					]}]},
				]} : "",
				{tag: "li", attrs: {"data-cmd": "reset"}, text: "\u2715"},
				(/^https?:$/.test(win.location.protocol) && cfg.sendTo.length
					? {tag: "li", attrs: {"class": "send_hosts"}, nodes: [
						"\u2197", {tag: "ul", nodes: cfg.sendTo.map(function(item) {
							item = item.split("|");
							return {tag: "li", nodes: [{tag: "a", attrs: {href: item.slice(1).join("|")}, text: item[0]}]};
						})}
					]}
					: null),
				media_type() === "video" ? "" : {tag: "li", attrs: {"data-cmd": "frames"}, text: "\u22EF"},
				{tag: "li", attrs: {"data-cmd": "options"}, text: "\u2713"},
			]);

			menu.style.cssText = "display: none; left: -" + menu.offsetWidth + "px";

			menu.addEventListener("mousedown", function(e) {
				var t = e.target;

				if (t.href && t.href.indexOf("%", t.host.length) > -1) {
					t.href = t.href.replace(/%url/, encodeURIComponent(media.src));
				}

				pdsp(e, !!t.textContent);
			}, false);

			// load favicons only when the menu item is hovered the first time
			if (/^https?:$/.test(win.location.protocol) && cfg.sendTo.length) {
				menu.querySelector(".send_hosts").onmouseover = function() {
					this.onmouseover = null;

					[].forEach.call(this.querySelectorAll(".send_hosts > ul > li > a"), function(a) {
						a.style.backgroundImage = "url(" + a.protocol + "//" + (a.getAttribute("data-favicon") || (a.host + "/favicon.ico")) + ")";
					});
				};
			}

			var handle_cmd = function(cmd, e) {
				if (e.button === 1) {
					return;
				}

				var p = e.type.indexOf("wheel") > -1 && (-e.wheelDelta || e.deltaY || e.detail) > 0 || e.button === 2 ? 1 : 0;

				if (cmd === "cycle") {
					media.cycle(!p);
				}
				else if (cmd === "flip") {
					media.flip(media, p);
				}
				else if (cmd === "rotate") {
					media.rotate(!p, e.ctrlKey);
				}
				else if (cmd === "zoom") {
					pdsp(e);
					media.zoomToCenter({wheelDelta: p ? -1 : 1});
				}
				else if (cmd === "reset" && e.button === 0) {
					media.reset();
				}
				else if (cmd === "frames") {
					Port.listen(function(js) {
						js = js && (js.data || js.message || js);
						Port.listen(null);
						Function("win", "draw_full_frame", "error_handler", js['frames.js'])(win, e.button === 0, function(msg) {
							// success
							if (msg === null) {
								if (browser.maxthon) {
									win.addEventListener("mousedown", function(e) {
										pdsp(e, false);
									}, true);
								}
							}
							else {
								alert(msg);
								msg = menu.querySelector("li[data-cmd='frames']");
								msg.removeAttribute("data-cmd");
								msg.style.opacity = "0.2";
							}
						});
					});

					Port.send({"cmd": "load_frames.js"});
				}
				else if (cmd === "options" && e.button === 0) {
					Port.send({"cmd": "open", "url": "options.html"});
				}
				else {
					p = null;
				}

				if (p !== null) {
					pdsp(e);
				}
			};

			menu.addEventListener(browser["wheel"], function(e) {
				var t = e.target;

				if (t.nodeType === 3) {
					t = t.parentNode;
				}

				if (t.type === "range") {
					t.value = Math.max(t.getAttribute("min"), Math.min(parseInt(t.value, 10) + t.getAttribute("step") * ((-e.wheelDelta || e.deltaY || e.detail) > 0 ? -1 : 1), t.getAttribute("max")));
					menu.onchange(e);
					pdsp(e);
				}
				else if (!/reset|frames|options/.test(t.getAttribute("data-cmd") || (t = t.parentNode) && t.getAttribute("data-cmd"))) {
					handle_cmd(t.getAttribute("data-cmd"), e);
				}
			}, false);

			menu.onclick = function(e) {
				var t = e.target;

				if (t.hasAttribute("data-cmd")
					|| (t = t.parentNode) && t.hasAttribute("data-cmd")) {
					handle_cmd(t.getAttribute("data-cmd"), e);
				}
			};

			menu.oncontextmenu = function(e) {
				var t = e.target;

				if (t.type === "range") {
					t.value = t.defaultValue;
					menu.onchange(e);
					pdsp(e);
				}
				else if (t.hasAttribute("data-cmd")
					|| (t = t.parentNode) && t.hasAttribute("data-cmd")) {
					handle_cmd(t.getAttribute("data-cmd"), e);
				}
			};

			menu.addEventListener(browser["transitionend"], function(e) {
				if (e.propertyName === "left" && this.style.left[0] === "-") {
					menu.style.display = "none";
				}
			}, false);

			if (win.Node.prototype && !win.Node.prototype.contains) {
				win.Node.prototype.contains = function(n) {
					return n instanceof Node && (this === n || !!(this.compareDocumentPosition(n) & 16));
				};
			}

			menu.onmouseover = function() {
				if (menu.mtimer) {
					clearTimeout(menu.mtimer);
					menu.mtimer = null;
					menu.style.left = "0";
					menu.style.opacity = "1";
				}
			};

			menu.onmouseout = function(e) {
				if (!this.contains(e.relatedTarget)) {
					doc.addEventListener("mousemove", menu_trigger, false);
					menu.mtimer = setTimeout(function() {
						menu.style.left = "-" + menu.offsetWidth + "px";
						menu.style.opacity = "0";
						menu.mtimer = null;
					}, 800);
				}
			};

			var menu_trigger = function(e) {
				if (!moving && !e.ctrlKey && e.clientX < 40 && e.clientY < win.innerHeight / 3) {
					if (menu.style.display !== "block") {
						menu.style.display = "block";

						setTimeout(function() {
							menu.style.left = "0";
							menu.style.opacity = "1";
						}, 50);

						doc.removeEventListener("mousemove", menu_trigger, false);
					}
				}
			};

			// Safari showed the menu even if the cursor wasn't at the edge
			setTimeout(function() {
				doc.addEventListener("mousemove", menu_trigger, false);
			}, 500);
		}

		after_calc = function() {
			if (media.calcFit) {
				winH = doc.compatMode[0] === "B" ? doc.body : root;
				winW = winH.clientWidth;
				winH = winH.clientHeight;

				media.setPos();

				setTimeout(function() {
					// Gecko and WebKit/Blink render scrollbars for a split of a second
					// even if the image fits into the view-port
					winH = doc.compatMode[0] === "B" ? doc.body : root;
					winW = winH.clientWidth;
					winH = winH.clientHeight;

					media.calcFit();
					media.setInfo();
					media.setCur();
				}, 0xf);
			}
		};

		media.resize = function(m, w) {
			if (m === -1 || (m === -1 && fZoom)) {
				if (w) {
					this.style.width = w;
				}
			}

			if (m === void 0) {
				m = 1;
			}

			if (m !== -1 && this.mode === -1) {
				this.style.width = this.style.height = "";
			}

			this.mode = m;
			this.className = "m_" + m;

			if (this.naturalWidth) {
				after_calc();
			}
		};

		media.setPos = function() {
			var s = this.style;

			if (cfg.center) {
				s.top = Math.max(0, (winH - this.offsetHeight) / 2) + "px";
				s.left = Math.max(0, (winW - this.offsetWidth) / 2) + "px";
			}
			else {
				s.top = s.left = "0";
			}

			this.box = this.getBoundingClientRect();

			if (this.box.left < 0) {
				s.left = (parseInt(s.left, 10) - this.box.left - win.pageXOffset) + "px";
			}

			if (this.box.top < 0) {
				s.top = (parseInt(s.top, 10) - this.box.top - win.pageYOffset) + "px";
			}
		};

		media.setInfo = function() {
			if (!cfg.imgInfo || !(media.naturalWidth && media.naturalHeight)) {
				return;
			}

			doc.title = cfg.imgInfo.replace(/%(o?[wh]|url|name|ratio|perc)/g, function(a, a1) {
				switch (a1) {
					case "w": return media.clientWidth;
					case "h": return media.clientHeight;
					case "ow": return media.naturalWidth;
					case "oh": return media.naturalHeight;
					case "url": return win.location.href;
					case "name": return media.alt;
					case "ratio":
						return Math.round(media.clientWidth / media.clientHeight * 100) / 100;
					case "perc":
						return ~~(0.5 + 1e4 * (media.clientWidth / media.naturalWidth)) / 100;
				}
			});
		};

		media.calcFit = function() {
			var box = this.box || this.getBoundingClientRect();

			noFit = {
				cur: (box.width <= winW && box.height <= winH),
				real: (this.naturalWidth <= winW && this.naturalHeight <= winH)
			};
		};

		media.wheelZoom = function(e) {
			stopScroll();
			pdsp(e);

			var w = media.offsetWidth, h = media.offsetHeight;

			if ((-e.wheelDelta || e.deltaY || e.detail) > 0) {
				media.resize(-1, Math.max(1, w * 0.75) + "px");
			}
			else {
				var width = w * (4 / 3);

				media.resize(-1, (width > 10 ? width : width + 3) + "px");
			}

			if (e.keypress || e.target.nodeName.toUpperCase() === "IMG") {
				win.scrollTo(
					(e.offsetX || e.layerX || 0) * media.offsetWidth / w - e.clientX + borderSize,
					(e.offsetY || e.layerY || 0) * media.offsetHeight / h - e.clientY + borderSize
				);
			}
		};

		media.zoomToCenter = function(e) {
			pdsp(e);
			this.wheelZoom({
				keypress: true,
				wheelDelta: e.wheelDelta,
				clientX: winW / 2,
				clientY: winH / 2,
				offsetX: win.pageXOffset + media.offsetLeft + winW / 2,
				offsetY: win.pageYOffset + media.offsetTop + winH / 2
			});
		};

		media.cycle = function(back) {
			var n_s = (this.mode === 1 ? 0 : this.mode) - (back ? 1 : -1);

			if (n_s === 1) {
				n_s = n_s - (back ? 1 : -1);
			}
			else if (n_s < 0) {
				n_s = 3;
			}
			else if (n_s > 3) {
				n_s = 0;
			}

			this.resize(n_s);
		};

		media.reset = function() {
			if (fZoom) {
				doc.removeEventListener("mousemove", draw_mask, false);
				doc.body.removeChild(media.mask);

				fZoom = null;
				cancelAction = false;
				return;
			}

			// Firefox document readyState stucks in "interactive" for videos,
			// but videos doesn't need to be loaded anyway
			if (media_type() === "img" && doc.readyState !== "complete") {
				return;
			}

			var filters = doc.body.querySelector("#menu li.filters > form");

			if (filters) {
				this.style.filter = this.style.webkitFilter = "";
				media.filters = {};
				filters.reset();
			}

			delete this.curdeg;
			delete this.scale;
			this.style[browser["transform"]] =
					this.style.width = this.style.height = "";
			this.resize(0);
		};

		media.flip = function(el, ori) {
			if (!this.scale) {
				this.scale = {"h": 1, "v": 1};
			}

			this.scale[ori ? "h" : "v"] *= -1;
			ori = this.scale.h !== 1 || this.scale.v !== 1 ?
				"scale(" + this.scale.h + "," + this.scale.v + ")" :
				"";

			if (this.curdeg) {
				ori += " rotate(" + this.curdeg + "deg)";
			}

			this.style[browser["transform"]] = ori;
		};

		media.rotate = function(deg, fine) {
			var rot;

			if (!this.curdeg) {
				this.curdeg = 0;
			}

			if (deg) {
				this.curdeg += (fine ? 10 : 90);
			}
			else {
				this.curdeg -= (fine ? 10 : 90);
			}

			win.status = this.curdeg + "°";

			rot = "rotate(" + this.curdeg + "deg)";

			if (this.scale) {
				rot += " scale(" + this.scale.h + ", " + this.scale.v + ")";
			}

			this.style[browser["transform"]] = rot;
			this.setPos();
		};

		if (cfg.wheelZoom) {
			doc.addEventListener(browser["wheel"], media.wheelZoom, false);
		}
		else {
			doc.addEventListener(browser["wheel"], function(e) {
				if (media.clientWidth > winW || media.clientHeight > winH) {
					stopScroll();

					var x = 0, y = ((-e.wheelDelta || e.deltaY || e.detail) > 0 ? winH : -winH) / 5;

					if (media.clientWidth <= winW && media.clientHeight > winH) {
						if (!cfg.hiddenScrollbars) {
							return;
						}
					}
					else if (media.clientHeight <= winH && media.clientWidth > winW
						|| e.clientX < winW / 2 && e.clientY > winH - 100) {
						x = (y < 0 ? -winW : winW) / 5;
						y = 0;
					}
					else if (!cfg.hiddenScrollbars) {
						return;
					}

					win.scrollBy(x, y);
					pdsp(e);
				}
			}, false);
		}

		media.setCur = function() {
			if (this.clientWidth > winW || this.clientHeight > winH) {
				this.style.cursor = "move";
			}
			else if (
				((winH >= this.clientHeight && winW >= this.clientWidth)
					&& (winH < this.naturalHeight || winW < this.naturalWidth))
				|| (noFit.cur && noFit.real
					&& (this.clientHeight < this.naturalHeight || this.clientWidth < this.naturalWidth))) {
				this.style.cursor = browser["zoom-in"];
			}
			else {
				this.style.cursor = "default";
			}
		};

		win.addEventListener("resize", after_calc, false);

		doc.addEventListener("contextmenu", function(e) {
			doc.removeEventListener("mousemove", onmove, true);

			if (progress) {
				clearTimeout(progress);
				progress = null;
			}

			if (cancelAction) {
				e.preventDefault();
			}
		}, false);

		var x, y;

		var onmove = function(e) {
			x = e.clientX;
			y = e.clientY;

			if (x === lastEvent.clientX && y === lastEvent.clientY) {
				return;
			}
			else if (progress) {
				clearTimeout(progress);
				progress = lastEvent.button = null;
			}

			if (sX === true || noFit.cur) {
				doc.removeEventListener("mousemove", onmove, true);
				return;
			}

			if (!moving) {
				moving = win.requestAnimationFrame(onmoveframe);
			}

			if (dragSlide.length !== 3) {
				dragSlide = [[x, y], [x, y], [x, y]];
				return;
			}

			// Opera fires move event before mouseup
			if (dragSlide[2][0] === x && dragSlide[2][1] === y) {
				return;
			}

			dragSlide[0] = [dragSlide[1][0], dragSlide[1][1]];
			dragSlide[1] = [dragSlide[2][0], dragSlide[2][1]];
			dragSlide[2] = [x, y];

			media.dragSlideTime = Date.now();
		};

		var pan = function() {
			win.scrollBy(sX - x, sY - y);

			sX = x;
			sY = y;
		};

		var onmoveframe = function() {
			pan();
			moving = null;
		};

		var stopScroll = function(e) {
			if (dragSlide.length) {
				clearInterval(progress);
				progress = null;
				cancelAction = false;
				dragSlide.length = 0;
				media.dragSlideTime = false;
			}

			if (e) {
				doc.removeEventListener(browser["wheel"], stopScroll, false);
			}
		};

		var startScroll = function() {
			progress = setInterval(function() {
				win.scrollBy(dragSlide[0], dragSlide[1]);

				if (dragSlide[0]) {
					if (Math.abs(dragSlide[0]) < 1) {
						dragSlide[0] = 0;
					}

					dragSlide[0] /= 1.11;
				}

				if (dragSlide[1]) {
					if (Math.abs(dragSlide[1]) < 1) {
						dragSlide[1] = 0;
					}

					dragSlide[1] /= 1.11;
				}

				var
					atRight = root.scrollWidth - win.pageXOffset === winW,
					atBottom = root.scrollHeight - win.pageYOffset === winH;

				if (!(dragSlide[0] && dragSlide[1])
						|| (!win.pageYOffset && !win.pageXOffset) || (!win.pageYOffset && atRight)
						|| (atBottom && !win.pageXOffset) || (atBottom && atRight)) {
					stopScroll();
				}
			}, 25);
		};

		var draw_mask = function(e) {
			if (++fZoom.counter % 3) {
				return;
			}

			var
				x = e.clientX - fZoom.left,
				y = e.clientY - fZoom.top;

			if (e.ctrlKey) {
				var
					rx = fZoom.prevX ? x - fZoom.prevX : 0,
					ry = fZoom.prevY ? y - fZoom.prevY : 0;

				fZoom.prevX = x;
				fZoom.prevY = y;

				fZoom.x += rx;
				fZoom.y += ry;
				fZoom.X += rx;
				fZoom.Y += ry;
			}
			else {
				if (fZoom.prevX) {
					delete fZoom.prevX;
					delete fZoom.prevY;
				}

				fZoom.w = Math.abs(fZoom.X - x);
				fZoom.h = Math.abs(fZoom.Y - y);
				fZoom.x = fZoom.X < x ? fZoom.X : fZoom.X - fZoom.w;
				fZoom.y = fZoom.Y < y ? fZoom.Y : fZoom.Y - fZoom.h;
			}

			media.mctx.clearRect(0, 0, media.mask.width, media.mask.height);
			media.mctx.fillRect(0, 0, media.mask.width, media.mask.height);
			media.mctx.clearRect(fZoom.x, fZoom.y, fZoom.w, fZoom.h);
		};

		var longpress_handler = function() {
			cancelAction = true;
			progress = null;

			longpress_func(cfg[lastEvent.button === 2 ? "lpRight" : "lpLeft"]);
		},
		longpress_func = function(id) {
			switch (id) {
				case 1: // fit to width/height
					var x, y;

					if (media.mode === 2 || media.clientWidth === doc.body.clientWidth
						|| (doc.body.clientHeight > media.clientHeight && winW < doc.body.clientWidth)) {
						x = media.clientWidth;
						media.resize(3);
						x = lastEvent.layerX * media.clientWidth / x - winW / 2;
						y = 0;
					}
					else {
						y = media.clientHeight;
						media.resize(2);
						x = 0;
						y = lastEvent.layerY * media.clientHeight / y - winH / 2;
					}

					win.scrollTo(x, y);
				break;
			}
		};

		media.addEventListener("mousedown", function(e) {
			pdsp(e, false);

			if (e.button === 1 || e.ctrlKey || e.altKey
				|| !e.shiftKey && media_type() === "video"
					&& (e.offsetY || e.layerY || 0) > Math.min(media.clientHeight - 40, media.clientHeight / 2)) {
				return;
			}

			if ((e.button === 0 && media.mode < 4) || e.button === 2) {
				if (e.button === 2) {
					sX = true;
				}
				else {
					if (!e.shiftKey && noFit.cur && media.clientWidth >= 60
						&& media.clientWidth - (e.offsetX || e.layerX || 0) <= 30) {
						lastEvent.clientX = e.clientX;
						lastEvent.clientY = e.clientY;
						return;
					}

					e.preventDefault();

					win.focus();
					sX = e.clientX;
					sY = e.clientY;
				}

				if (!e.shiftKey) {
					doc.addEventListener("mousemove", onmove, true);
				}
			}

			if (!cfg.lpDelay) {
				return;
			}

			// for fine move and free zoom
			if (e.shiftKey) {
				cancelAction = true;

				if (e.button === 0) {
					fZoom = {
						counter: 0,
						left: parseInt(media.style.left, 10),
						top: parseInt(media.style.top, 10)
					};

					fZoom.X = e.clientX - fZoom.left;
					fZoom.Y = e.clientY - fZoom.top;

					if (!media.mask) {
						media.mask = doc.createElement("canvas");
						media.mask.className = "mask";
						media.mask.style.cssText = "display: block; position: fixed; left: 0; top: 0;";
						media.mctx = media.mask.getContext("2d");
					}

					doc.addEventListener("mousemove", draw_mask, false);

					var h = media.curdeg && Math.sin(media.curdeg) ? true : false,
						w = h ? media.offsetHeight : media.offsetWidth;
					h = h ? media.offsetWidth : media.offsetHeight;
					media.mask.width = Math.min(winW, w);
					media.mask.height = Math.min(winH, h);

					media.mask.style.left = media.style.left;
					media.mask.style.top = media.style.top;

					media.mctx.clearRect(0, 0, media.mask.width, media.mask.height);
					doc.body.appendChild(media.mask);

					w = win.getComputedStyle(media.mask).color;
					media.mctx.fillStyle = !w || w === "rgb(0, 0, 0)" ? "rgba(0,0,0,.4)" : w;
				}
			}

			// is dragSlideing
			if (progress) {
				if (stopScroll) {
					stopScroll();
					cancelAction = e.button === 0;
				}

				if (e.button === 0) {
					return;
				}
			}

			if (e.shiftKey || (e.button === 0 && !cfg.lpLeft)
				|| (e.button === 2 && !cfg.lpRight)) {
				return;
			}

			lastEvent.clientX = e.clientX;
			lastEvent.clientY = e.clientY;
			lastEvent.layerX = e.offsetX || e.layerX || 0;
			lastEvent.layerY = e.offsetY || e.layerY || 0;
			lastEvent.button = e.button;

			progress = setTimeout(longpress_handler, cfg.lpDelay);
		}, true);

		doc.addEventListener("mouseup", function(e) {
			if (e.button !== 0) {
				return;
			}

			var x, y;

			if (media.mode < 4) {
				doc.removeEventListener("mousemove", onmove, true);

				if (fZoom) {
					doc.removeEventListener("mousemove", draw_mask, false);
					doc.body.removeChild(media.mask);

					var
						w = Math.min(
							media.clientWidth,
							fZoom.w + Math.min(fZoom.x, 0)
								+ (fZoom.x + fZoom.w > media.mask.width ?
										media.mask.width - fZoom.x - fZoom.w : 0)
						),
						h = Math.min(
							media.clientHeight,
							fZoom.h + Math.min(fZoom.y, 0)
								+ (fZoom.y + fZoom.h > media.mask.height ?
										media.mask.height - fZoom.y - fZoom.h : 0)
						);

					x = Math.max(0, fZoom.x);
					y = Math.max(0, fZoom.y);

					fZoom = null;
					cancelAction = false;

					if (!w || !h || x >= media.clientWidth || y >= media.clientHeight) {
						return;
					}

					var nimgw, nimgh,
						scrollbars = (win.innerWidth - winW) || (win.innerHeight - winH);

					var fitW = winW < w * winH / h;

					fitW = e.ctrlKey ? !fitW : fitW;

					if (fitW) {
						nimgw = media.clientWidth * winW / w;
						nimgh = nimgw * media.clientHeight / media.clientWidth;
					}
					else {
						nimgh = media.clientHeight * winH / h;
						nimgw = nimgh * media.clientWidth / media.clientHeight;
					}

					if (nimgw > MAXSIZE) {
						nimgw = MAXSIZE;
						nimgh = media.clientHeight * MAXSIZE / media.clientWidth;
					}

					if (nimgh > MAXSIZE) {
						nimgw = media.clientWidth * MAXSIZE / media.clientHeight;
						nimgh = MAXSIZE;
					}

					var cx = (win.pageXOffset + x + w / 2) * nimgw /
										media.clientWidth - winW / 2 - scrollbars;
					var cy = (win.pageYOffset + y + h / 2) * nimgh /
										media.clientHeight - winH / 2 - scrollbars;

					media.resize(-1, nimgw + "px");

					win.scrollTo(cx, cy);
					return;
				}
				else if (dragSlide.length === 3) {
					x = dragSlide[0][0] - dragSlide[2][0];
					y = dragSlide[0][1] - dragSlide[2][1];

					if ((Date.now() - media.dragSlideTime) > 100) {
						cancelAction = false;
						dragSlide.length = 0;
						return;
					}

					if (x || y) {
						dragSlide.length = 2;
						dragSlide[0] = x;
						dragSlide[1] = y;
						startScroll();
						doc.addEventListener(browser["wheel"], stopScroll, false);
						return;
					}
				}
			}

			if (progress) {
				clearTimeout(progress);
				progress = null;
			}

			if (cancelAction) {
				cancelAction = false;
				return;
			}

			if (e.shiftKey || e.ctrlKey || e.altKey
				|| e.target !== media || lastEvent.button === null
				|| e.clientX !== lastEvent.clientX || e.clientY !== lastEvent.clientY) {
				return;
			}

			if (media.mode < 2 && noFit.real) {
				if (media.mode === -1) {
					media.resize(0);
				}
				else if (winW / winH < media.naturalWidth / media.naturalHeight
					|| winH === media.offsetHeight && winW > media.offsetWidth) {
					media.resize(2);
				}
				else {
					media.resize(3);
				}
			}
			else if (media.mode === 1 || noFit.cur) {
				if (media.naturalWidth === media.clientWidth
					&& media.naturalHeight === media.clientHeight) {
					return;
				}

				x = e.offsetX || e.layerX || 0;
				y = e.offsetY || e.layerY || 0;

				if (media.scale) {
					if (media.scale.h === -1) {
						x = media.clientWidth - x;
					}

					if (media.scale.v === -1) {
						y = media.clientHeight - y;
					}
				}

				x = x * media.naturalWidth / media.clientWidth - winW / 2;
				y = y * media.naturalHeight / media.clientHeight - winH / 2;

				media.resize(0);
				win.scrollTo(x, y);
			}
			else {
				media.resize(1);
			}
		}, true);

		doc.addEventListener("keydown", function(e) {
			if (stopScroll) {
				stopScroll();
			}

			if (shortcut.isModifier(e)) {
				return;
			}

			var key = shortcut.key(e);

			if (e.ctrlKey || (fZoom && key !== "Esc")) {
				return;
			}

			if (key === "+" || key === "-") {
				e.wheelDelta = key === "+" ? 1 : -1;
				media.zoomToCenter(e);
				return false;
			}

			var x, y, z;

			switch (key) {
				case "Esc":
					media.reset();
				break;

				case "PgUp":
					x = e.shiftKey ? -winW / 2 : 0;
					y = e.shiftKey ? 0 : -winH / 2;
				break;

				case "PgDn":
					x = e.shiftKey ? winW / 2 : 0;
					y = e.shiftKey ? 0 : winH / 2;
				break;

				case "End":
					x = e.shiftKey ? root.scrollWidth : win.pageXOffset;
					y = e.shiftKey ? win.pageYOffset : root.scrollHeight;
					z = true;
				break;

				case "Home":
					x = e.shiftKey ? 0 : win.pageXOffset;
					y = e.shiftKey ? win.pageYOffset : 0;
					z = true;
				break;

				default: x = true;
			}

			if (x !== true) {
				if (x !== void 0) {
					win[z ? "scrollTo" : "scrollBy"](x, y);
					e.preventDefault();
				}

				return;
			}

			switch (key) {
				case cfg.key_mOrig: media.resize(0); break;
				case cfg.key_mFit: media.resize(1); break;
				case cfg.key_mFitW: media.resize(2); break;
				case cfg.key_mFitH: media.resize(3); break;
				case cfg.key_cycle: media.cycle(e.shiftKey); break;
				case cfg.key_rotL: media.rotate(false, e.shiftKey); break;
				case cfg.key_rotR: media.rotate(true, e.shiftKey); break;
				case cfg.key_flipH: media.flip(media, 0); break;
				case cfg.key_flipV: media.flip(media, 1); break;
				default: x = true;
			}

			if (media_type() === "video") {
				x = null;

				switch (key) {
					case "Space": media.togglePlay(); break;
					case "Up": case "Down":
						key = media.volume + (key === "Up" ? 0.1 : -0.1);
						// must check the range manually, otherwise browsers will throw an error
						media.volume = key < 0 ? 0 : (key > 1 ? 1 : key);
						break;
					default: x = true;
				}
			}

			if (x !== true) {
				pdsp(e);
			}
		}, true);

		if (media_type() === "video") {
			media.togglePlay = function() {
				if (this.paused || Math.abs(this.duration - this.currentTime) < 0.01) {
					this.play();
				}
				else {
					this.pause();
				}
			};

			media.addEventListener("click", function(e) {
				var top = (e.offsetY || e.layerY || 0);

				if (top > media.clientHeight - 40) {
					return;
				}

				if (top > media.clientHeight / 2) {
					media.togglePlay();
				}

				pdsp(e);
			});

			media.addEventListener("dblclick", function(e) {
				if ((e.offsetY || e.layerY || 0) < media.clientHeight / 2) {
					pdsp(e);
				}
			});
		}

		media.mode = cfg.mode;
		media.calcFit();
		progress = [];

		if (cfg.minUpscale) {
			if (media.naturalWidth >= winW * cfg.minUpscale / 100) {
				progress[0] = 1;
			}

			if (media.naturalHeight >= winH * cfg.minUpscale / 100) {
				progress[1] = 1;
			}
		}

		if (media.mode === 4) {
			media.mode = media.naturalWidth / media.naturalHeight > winW / winH ? 3 : 2;
		}
		else if ((media.mode === 1 || (media.mode === 0 && noFit.real))
			&& progress.length) {
			media.mode = media.naturalWidth / media.naturalHeight > winW / winH ? 2 : 3;
		}

		if ((media.mode === 2 && media.naturalWidth < winW && !progress[0])
			|| (media.mode === 3 && media.naturalHeight < winH && !progress[1])) {
			media.mode = 0;
		}

		progress = null;
		media.resize(media.mode);
	};

	if (media_type() === "video") {
		root = {};
		cfg.vidattrs.split(/\s+/).forEach(function(attr) {
			attr = attr.split("=");

			if (attr[0] === "volume") {
				root[attr[0]] = Math.min(100, Math.max(0, parseInt(attr[1]) / 100));
			}
			else {
				root[attr[0]] = true;
			}
		});

		["autoplay", "loop", "controls", "muted", "volume"].forEach(function(attr) {
			media[attr] = attr === "volume" ? (root[attr] === undefined ? 1 : root[attr]) : root[attr] || false;
		});

		if (media.autoplay) {
			media.onloadedmetadata = function() {
				this.onloadedmetadata = null;

				setTimeout(function() {
					media.currentTime = 0;
					media.play();
				}, 100);
			};
		}
	}
	else {
		media = doc.createElement("img");
		media.src = win.location.href;
	}

	root = doc.documentElement;

	media.onerror = function() {
		// Opera fires an error event on local video files when they're close to the end
		// Firefox did it a few times too, however there was nothing wrong with the video
		if (media.currentTime > 0 && win.location.protocol === "file:") {
			return;
		}

		clearInterval(progress);
		root.classList.add("error");
	};

	if (/^data:/.test(win.location.href)) {
		media.alt = media_type() + " (data:)";

		if (!cfg.imgInfo) {
			doc.title = media.alt;
		}
	}
	else {
		media.alt = win.location.href.replace(/#.*/, "").match(/(?:[^\/]+)?$/)[0] || media_type();
		media.alt = media.alt.split("?")[0];

		try {
			// some Unicode characters caused problems
			media.alt = decodeURIComponent(media.alt);
		} catch (ex) {}
	}

	if (media.naturalWidth || media.videoWidth) {
		init_params = null;
		// Firefox and Safari won't position the image without setTimeout
		setTimeout(init, 0xf);
	}
	else {
		init_params = {loop: 0, wait: 150, maxLoop: 900};
		progress = setInterval(init, init_params.wait);

		setTimeout(function() {
			if (root.classList.contains(media_type())) {
				return;
			}

			// Firefox won't show the image if a site uses CSP settings,
			// so we should use the default image
			if ((browser.firefox || browser.maxthon)
				&& (media_type() === "img"
					? !media.naturalWidth && doc.images[0].naturalWidth
					: !media.videoWidth && doc.body.querySelector("video").videoWidth)) {
				doc.images[0].style.display = "block";
				clearInterval(progress);
				return;
			}

			media.style.display = "none";
			var h2 = doc.body.appendChild(doc.createElement("h2"));
			h2.textContent = media.alt;
			h2.onclick = function() {
				doc.body.removeChild(this);
				media.style.display = "block";
			};
		}, 1000);
	}
};

(function() {
	if (browser.safari && location.protocol === "safari-extension:") {
		init = null;
		return;
	}

	if (browser.opera || browser.firefox) {
		if (media_type()) {
			Port.listen(function(e) {
				Port.listen(null);
				init(window, document, e);
			});
			Port.send({"cmd": "hello"});
		}
		else {
			init = null;
		}
	}
	else {
		Port.listen(function(e) {
			Port.listen(null);

			if (!media_type()) {
				init = null;
				return;
			}

			init(window, document, e);
		});

		Port.send({"cmd": "hello"});
	}
})();
